var Promise = require('when/es6-shim/Promise');
var underscore = require('underscore')
var EventEmitter = require('events').EventEmitter;

function setJobCount(num) {
	this._config.max_job = num;
	reloadJob(this)
}

function registerFunction(name, func) {
	if (this._callback[name]) {
		throw new Error("handler for " + name + " is already defined")
	}
	if (!underscore.isFunction(func)) {
		throw new Error("func is not a function")
	}
	this._callback[name] = func
}

function registerAction(name, argList, initDelay) {
	if (!this._callback[name]) {
		throw new Error("handler for " + name + " is not defined")
	}
	if (!argList) {
		argList = []
	}
	var that = this;
	var pr = new Promise(function(resolve, reject) {
		that._job.push({
			name: name,
			data: argList,
			resolve: resolve,
			reject: reject,
			delay: initDelay
		})
	})
	execJob(this);
	return pr;
}

function reloadJob(ctx){
	var job = ctx._job.slice(0);
	ctx._job = [];
	underscore.each(job,function(itemJob){
		ctx._job.push(itemJob)
		execJob(ctx)
	})
}

function up(ctx) {
	ctx._current_job += 1;
}

function down(ctx) {

	console.log("Remaining job count",ctx._job.length)
	ctx._current_job -= 1;
	if(ctx._job.length == 0){
		ctx.emit('empty')
	}
	execJob(ctx)
}

function dropQueue(){
	this._job = []
}

function execJob(ctx) {
	var execCount = ctx._current_job;
	var maxJobCount = ctx._config.max_job;
	if (execCount < maxJobCount && ctx._job.length) {
		up(ctx);
		var job = ctx._job.shift();

		var funcName = job.name;
		var funcArgs = job.data;
		var resolve = job.resolve;
		var reject = job.reject;
		var delay = job.delay;

		var funcHandler = ctx._callback[funcName];
		var ret = funcHandler.apply(null, funcArgs);
		//is thenable
		if (typeof ret.then == "function") {
			ret.then(function(result) {
				resolve(result)
				down(ctx)
			}, function(reason) {
				reject(reason)
				down(ctx)
			}).catch(function(err) {
				reject(err)
				down(ctx)
			})
		} else {
			delay = delay ? delay : 0;
			setTimeout(function() {
				resolve(ret);
				down(ctx)
			}, delay)
		}
	}
	// execJob(ctx)
}
function Throttle(num){
	underscore.extend(this,new EventEmitter)	
	this._callback = {}
	this._job = []
	this._config = {
		max_job:10
	}
	this._current_job = 0;
	this.setJobCount(num ? num : 10);
}

Throttle.prototype.setJobCount = setJobCount
Throttle.prototype.registerFunction = registerFunction
Throttle.prototype.registerAction = registerAction
Throttle.prototype.dropQueue = dropQueue;

module.exports = Throttle;
